// N.B. This file is not used by this usdt_sample. Instead, the StaticTracepoint.h file from folly is used.
//  It is here only for demonstration purposes.

/* Generated by the Systemtap dtrace wrapper */


#define _SDT_HAS_SEMAPHORES 1


#define STAP_HAS_SEMAPHORES 1 /* deprecated */


#include <sys/sdt.h>

/* USDT_SAMPLE_LIB1_OPERATION_START ( uint64_t operation_id, const char * input ) */
#if defined STAP_SDT_V1
#define USDT_SAMPLE_LIB1_OPERATION_START_ENABLED() __builtin_expect (operation_start_semaphore, 0)
#define usdt_sample_lib1_operation_start_semaphore operation_start_semaphore
#else
#define USDT_SAMPLE_LIB1_OPERATION_START_ENABLED() __builtin_expect (usdt_sample_lib1_operation_start_semaphore, 0)
#endif
__extension__ extern unsigned short usdt_sample_lib1_operation_start_semaphore __attribute__ ((unused)) __attribute__ ((section (".probes")));
#define USDT_SAMPLE_LIB1_OPERATION_START(arg1, arg2) \
DTRACE_PROBE2 (usdt_sample_lib1, operation_start, arg1, arg2)

/* USDT_SAMPLE_LIB1_OPERATION_END ( uint64_t operation_id, const char * output ) */
#if defined STAP_SDT_V1
#define USDT_SAMPLE_LIB1_OPERATION_END_ENABLED() __builtin_expect (operation_end_semaphore, 0)
#define usdt_sample_lib1_operation_end_semaphore operation_end_semaphore
#else
#define USDT_SAMPLE_LIB1_OPERATION_END_ENABLED() __builtin_expect (usdt_sample_lib1_operation_end_semaphore, 0)
#endif
__extension__ extern unsigned short usdt_sample_lib1_operation_end_semaphore __attribute__ ((unused)) __attribute__ ((section (".probes")));
#define USDT_SAMPLE_LIB1_OPERATION_END(arg1, arg2) \
DTRACE_PROBE2 (usdt_sample_lib1, operation_end, arg1, arg2)

