R"********(
/*
 * Copyright (c) 2018 Clevernet, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef BPF_LICENSE
/* No license defined, using GPL
 * You can define your own BPF_LICENSE in your C code */
#define BPF_LICENSE GPL
#endif
#define ___LICENSE(s) #s
#define __LICENSE(s) ___LICENSE(s)
#define _LICENSE __LICENSE(BPF_LICENSE)
char _license[] BCC_SEC("license") = _LICENSE;

)********"
